/*
Copyright (C) 2001 Lkb

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef DAEDALUS_COMBINERD3DRENDER_SECONDPART
#include <pshpack1.h>

struct TextureCoords2D
{
	union
	{
		D3DVALUE uv[2];
		struct
		{
			D3DVALUE u;
			D3DVALUE v;
		};
	};
};
#include <poppack.h>

// see <Microsoft Windows 2000/Me DDK>\src\win_me\display\mini\d3dref\inc\refrast.hpp, 739 for more info on vertex format

struct PVertex
{
protected:
	void* p;
	BYTE textures;
	bool diffuse;
	bool specular;

public:
	PVertex(void* p_p = 0, bool p_diffuse = false, bool p_specular = false, int p_textures = 0)
		: p(p_p), textures(p_textures), diffuse(p_diffuse), specular(p_specular)
	{}

	UINT Size()
	{
		return sizeof(D3DXVECTOR3) + sizeof(float) + (diffuse ? sizeof(D3DCOLOR) : 0) + (specular ? sizeof(D3DCOLOR) : 0) + (textures * sizeof(TextureCoords2D));
	}

	DWORD FVF() const
	{
		return (D3DFVF_XYZRHW | (textures << D3DFVF_TEXCOUNT_SHIFT) | (diffuse ? D3DFVF_DIFFUSE : 0) | (specular ? D3DFVF_SPECULAR : 0));
	}

	bool HasDiffuse() const
	{
		return diffuse;
	}

	bool HasSpecular() const
	{
		return specular;
	}

	BYTE GetTextureCordinatesNumber() const
	{
		return textures;
	}

	void*& Ptr()
	{
		return p;
	}

	D3DXVECTOR3& Position()
	{
		return *(D3DXVECTOR3*)p;
	}

	float& RHW()
	{
		return *(float*)((UINT_PTR)p + sizeof(D3DXVECTOR3));
	}

	D3DCOLOR& Diffuse()
	{
		return *(D3DCOLOR*)((UINT_PTR)p + sizeof(D3DXVECTOR3) + sizeof(float));
	}

	D3DCOLOR& Specular()
	{
		return *(D3DCOLOR*)((UINT_PTR)p + sizeof(D3DXVECTOR3) + sizeof(float) + (diffuse ? sizeof(D3DCOLOR) : 0));
	}

	TextureCoords2D& TextureCoords(int num)
	{
		return ((TextureCoords2D*)((UINT_PTR)p + sizeof(D3DXVECTOR3) + sizeof(float) + (diffuse ? sizeof(D3DCOLOR) : 0) + (specular ? sizeof(D3DCOLOR) : 0)))[num];
	}

	D3DVALUE& TextureU(int num)
	{
		return TextureCoords(num).u;
	}

	D3DVALUE& TextureV(int num)
	{
		return TextureCoords(num).v;
	}

	void* const& Ptr() const
	{
		return p;
	}

	const D3DXVECTOR3& Position() const
	{
		return *(D3DXVECTOR3*)p;
	}

	const float& RHW() const
	{
		return *(float*)((UINT_PTR)p + sizeof(D3DXVECTOR3));
	}

	const D3DCOLOR& Diffuse() const
	{
		return *(D3DCOLOR*)((UINT_PTR)p + sizeof(D3DXVECTOR3) + sizeof(float));
	}

	const D3DCOLOR& Specular() const
	{
		return *(D3DCOLOR*)((UINT_PTR)p + sizeof(D3DXVECTOR3) + sizeof(float) + (diffuse ? sizeof(D3DCOLOR) : 0));
	}

	const TextureCoords2D& TextureCoords(int num) const
	{
		return ((TextureCoords2D*)((UINT_PTR)p + sizeof(D3DXVECTOR3) + sizeof(float) + (diffuse ? sizeof(D3DCOLOR) : 0) + (specular ? sizeof(D3DCOLOR) : 0)))[num];
	}

	const D3DVALUE& TextureU(int num) const
	{
		return TextureCoords(num).u;
	}

	const D3DVALUE& TextureV(int num) const
	{
		return TextureCoords(num).v;
	}

	PVertex& operator +=(UINT_PTR value)
	{
		p = (void*)((UINT_PTR)p + (Size() * value));
		return *this;
	}

	PVertex& operator -=(UINT_PTR value)
	{
		p = (void*)((UINT_PTR)p - (Size() * value));
		return *this;
	}

	void Set(PVertex src)
	{
		Position() = src.Position();
		RHW() = src.RHW();

		if(HasDiffuse() && src.HasDiffuse())
		{
			Diffuse() = src.Diffuse();
		}

		if(HasSpecular() && src.HasSpecular())
		{
			Specular() = src.Specular();
		}

		if(src.GetTextureCordinatesNumber() > 0)
		{
			int texToCopy = GetTextureCordinatesNumber();
			for(int iTex = 0; iTex < texToCopy; iTex++)
			{
				TextureCoords(iTex) = src.TextureCoords(iTex % src.GetTextureCordinatesNumber());
			}
		}
	}
};

struct CombinerPass;
struct CombinerInfo;

enum TextureType
{
	TextureType_Unused = 0,
	TextureType_SingleColor = 1,
	TextureType_Texture0 = 2,
	TextureType_Texture1 = 3 // for multitexturing
};

struct CombinerVertexModificationInfo
{
	D3DCOLOR diffuse;
	D3DCOLOR specular;
	bool overrideDiffuseColor;
	bool overrideDiffuseAlpha;
	bool overrideSpecularColor;
	bool overrideSpecularAlpha;
	TextureType textureTypes[8];
	UINT texturesCount;
};
#else

inline void D3DRender::ApplyDiffuseColor(const CombinerVertexModificationInfo& cds, PVertex p)
{
	if(cds.overrideDiffuseAlpha)
	{
		if(cds.overrideDiffuseColor)
		{
			p.Diffuse() = cds.diffuse;
		}
		else
		{
			D3DCOLOR orValue = cds.diffuse & 0xff000000;
			p.Diffuse() = (p.Diffuse() & 0x00ffffff) | orValue;
		}
	}
	else if(cds.overrideDiffuseColor)
	{
		D3DCOLOR orValue = cds.diffuse & 0x00ffffff;
		p.Diffuse() = (p.Diffuse() & 0xff000000) | orValue;
	}
}

inline void D3DRender::ApplySpecularColor(const CombinerVertexModificationInfo& cds, PVertex p)
{
	if(cds.overrideSpecularAlpha)
	{
		if(cds.overrideSpecularColor)
		{
			p.Specular() = cds.specular;
		}
		else
		{
			D3DCOLOR orValue = cds.specular & 0xff000000;
			p.Specular() = (p.Specular() & 0x00ffffff) | orValue;
		}
	}
	else if(cds.overrideSpecularColor)
	{
		D3DCOLOR orValue = cds.specular & 0x00ffffff;
		p.Specular() = (p.Specular() & 0xff000000) | orValue;
	}
}

inline void D3DRender::ProcessVertices(const CombinerVertexModificationInfo& cds, PVertex& p, int count)
{
	PVertex ivout(
		m_ucTempVertexBuffer,
		(cds.overrideDiffuseColor || cds.overrideDiffuseAlpha || p.HasDiffuse()),
		(cds.overrideSpecularColor || cds.overrideSpecularAlpha || p.HasSpecular()),
		cds.texturesCount
		);
	
	PVertex ivin = p;
	p = ivout;
	for(int i = 0; i < count; i++)
	{
		ivout.Set(ivin);
		ApplyDiffuseColor(cds, ivout);
		ApplySpecularColor(cds, ivout);
		for(int iTex = 0; iTex < ivout.GetTextureCordinatesNumber(); iTex++)
		{
			if(cds.textureTypes[iTex] == TextureType_SingleColor)
			{
				// correct?
				ivout.TextureU(iTex) = 0.0f;
				ivout.TextureV(iTex) = 0.0f;
			}
		}
		
		ivin += 1;
		ivout += 1;
	}
}

#endif
